@@ -18,7 +18,7 @@ class CWLJobGatherer(BaseOperator):
         task_id,
         *args, **kwargs
     ):
-        super().__init__(task_id=task_id, *args, **kwargs)
+        super().__init__(task_id=task_id, trigger_rule="none_failed", *args, **kwargs)  # change default trigger_rule as the upstream can be skipped
         
 
     def execute(self, context):
@@ -33,7 +33,7 @@ class CWLJobGatherer(BaseOperator):
         _, workflow_report = relocate_outputs(
             workflow=context["dag"].workflow,
             job_data=collect_reports(context),
-            cwl_args=context["dag"].default_args["cwl"],
+            cwl_args=context["dag"].default_args["cwl"]
         )
      
         return workflow_report
