@@ -38,6 +38,7 @@ from cwltool.context import (
     RuntimeContext
 )
 from cwltool.process import relocateOutputs
+from cwltool.expression import do_eval
 from cwltool.load_tool import (
     load_tool,
     jobloaderctx
@@ -58,7 +59,8 @@ from cwl_airflow.utilities.helpers import (
     remove_field_from_dict,
     get_uncompressed,
     get_compressed,
-    get_files
+    get_files,
+    get_dir_size
 )
 
 
@@ -67,12 +69,13 @@ CWL_OUTPUTS_FOLDER = os.path.join(AIRFLOW_HOME, "cwl_outputs_folder")
 CWL_INPUTS_FOLDER = os.path.join(AIRFLOW_HOME, "cwl_inputs_folder")
 CWL_PICKLE_FOLDER = os.path.join(AIRFLOW_HOME, "cwl_pickle_folder")
 CWL_USE_CONTAINER = True
+CWL_KEEP_TMP_DATA = False
 CWL_NO_MATCH_USER = False
 CWL_SKIP_SCHEMAS = True
 CWL_STRICT = False
 CWL_QUIET = False
-CWL_RM_TMPDIR = True       # better not to change without need
-CWL_MOVE_OUTPUTS = "move"  # better not to change without need
+CWL_RM_TMPDIR = True       # better not to change without need, however if we "copy", this will be always False, and that's good for disk usage statistics
+CWL_MOVE_OUTPUTS = "copy"  # we want to "copy" instead of "move" to calculate disk usage from the dag callback
 CWL_ENABLE_DEV = True      # better not to change without need
 
 
@@ -85,6 +88,36 @@ dag = CWLDAG(
 """
 
 
+def get_workflow_execution_stats(context):
+    job_data = collect_reports(context)
+    dag_run = context["dag_run"]
+    statistics = ""
+    try:
+        statistics = {
+            "version": "1.0",  # upgrade the version when we decide to change the logic of how we collect statistics
+            "total": {
+                "tmp_folder_size": get_dir_size(job_data["tmp_folder"]),
+                "outputs_folder_size": get_dir_size(job_data["outputs_folder"]),
+                "start_date": dag_run.start_date.isoformat(),
+                "end_date": round((dag_run.end_date - dag_run.start_date).total_seconds(), 3)
+            },
+            "steps": {}
+        }
+        for ti in context["dag_run"].get_task_instances():
+            step_tmp_folder, _, _, _ = get_temp_folders(
+                task_id=ti.task_id,
+                job_data=job_data
+            )
+            statistics["steps"][ti.task_id] = {
+                "tmp_folder_size": get_dir_size(step_tmp_folder),
+                "start_date": round((ti.start_date - dag_run.start_date).total_seconds(), 3),
+                "end_date": round((ti.end_date - dag_run.start_date).total_seconds(), 3)
+            }
+    except Exception as err:
+        logging.info(f"Failed to collect statistics for dag_id: {dag_run.dag_id}, run_id: {dag_run.run_id}, due to\n {err}")
+    return statistics
+
+
 def overwrite_deprecated_dag(
     dag_location,
     deprecated_dags_folder=None
@@ -237,7 +270,7 @@ def remove_dag_run_tmp_data(dag_run):
             report_location = ti.xcom_pull(task_ids=ti.task_id)
             tmp_folder_set.add(load_yaml(report_location)["tmp_folder"])
         except Exception:
-            logging.debug(" - report file has been already deleted or it's missing tmp_folder field")
+            logging.debug(" - XCom was empty, or report file has been already deleted, or it's missing tmp_folder field")
     for tmp_folder in tmp_folder_set:
         try:
             logging.info(f"Removing tmp data from {tmp_folder}")
@@ -297,7 +330,8 @@ def clean_up_dag_run(dag_id, run_id, dags_folder=None, kill_timeout=None):
 def conf_get(
     section,
     key,
-    default
+    default,
+    as_boolean=None
 ):
     """
     Return value from AirflowConfigParser object.
@@ -305,9 +339,13 @@ def conf_get(
     Suppresses annoying warning messages
     """
 
+    as_boolean = False if as_boolean is None else as_boolean
     try:
         logging.disable(logging.WARNING)
-        return conf.get(section, key)
+        if as_boolean:
+            return conf.getboolean(section, key)
+        else:
+            return conf.get(section, key)
     except AirflowConfigException:
         return default
     finally:
@@ -383,25 +421,35 @@ def get_default_cwl_args(preset_cwl_args=None):
                     preset_cwl_args.get("pickle_folder", CWL_PICKLE_FOLDER)
                 )
             ),
-            "use_container": conf_get(
+            "keep_tmp_data": conf_get(                                             # prevents from cleaning dag_run temp data and XCom's in both success or failure scenarios
+                "cwl", "keep_tmp_data",
+                preset_cwl_args.get("keep_tmp_data", CWL_KEEP_TMP_DATA),
+                True                                                               # return Boolean
+            ),
+            "use_container": conf_get(                                             # execute jobs in docker containers
                 "cwl", "use_container",
-                preset_cwl_args.get("use_container", CWL_USE_CONTAINER)            # execute jobs in docker containers
+                preset_cwl_args.get("use_container", CWL_USE_CONTAINER),
+                True                                                               # return Boolean
             ),
-            "no_match_user": conf_get(
+            "no_match_user": conf_get(                                             # disables passing the current uid to "docker run --user"
                 "cwl", "no_match_user",
-                preset_cwl_args.get("no_match_user", CWL_NO_MATCH_USER)            # disables passing the current uid to "docker run --user"
+                preset_cwl_args.get("no_match_user", CWL_NO_MATCH_USER),
+                True                                                               # return Boolean
             ),
-            "skip_schemas": conf_get(
+            "skip_schemas": conf_get(                                              # it looks like this doesn't influence anything in the latest cwltool
                 "cwl", "skip_schemas", 
-                preset_cwl_args.get("skip_schemas", CWL_SKIP_SCHEMAS)              # it looks like this doesn't influence anything in the latest cwltool
+                preset_cwl_args.get("skip_schemas", CWL_SKIP_SCHEMAS),
+                True                                                               # return Boolean
             ),
             "strict": conf_get(
                 "cwl", "strict", 
-                preset_cwl_args.get("strict", CWL_STRICT)
+                preset_cwl_args.get("strict", CWL_STRICT),
+                True                                                               # return Boolean
             ),
             "quiet": conf_get(
                 "cwl", "quiet", 
-                preset_cwl_args.get("quiet", CWL_QUIET)
+                preset_cwl_args.get("quiet", CWL_QUIET),
+                True                                                               # return Boolean
             ),
             "rm_tmpdir": preset_cwl_args.get("rm_tmpdir", CWL_RM_TMPDIR),          # even if we can set it in "preset_cwl_args" it's better not to change
             "move_outputs": preset_cwl_args.get("move_outputs", CWL_MOVE_OUTPUTS), # even if we can set it in "preset_cwl_args" it's better not to change
@@ -415,23 +463,21 @@ def get_default_cwl_args(preset_cwl_args=None):
 def relocate_outputs(
     workflow,
     job_data,
-    cwl_args=None,
-    remove_tmp_folder=None
+    cwl_args=None
 ):
     """
-    Relocates filtered outputs to "outputs_folder" and, by default,
-    removes tmp_folder, unless "remove_tmp_folder" is set to something
-    else. Saves report with relocated outputs as "workflow_report.json"
-    to "outputs_folder".
-    Maps outputs from "workflow" back to normal (from step_id_step_out
-    to workflow output) and filters "job_data" based on them (combining
-    items from "job_data" into a list based on "outputSource" if it
-    was a list). "cwl_args" can be used to update default parameters
-    used for loading and runtime contexts.
+    Moves or copies filtered outputs to "outputs_folder" depending on
+    "runtime_context.move_outputs" value, however "tmp_folder" is not
+    going to be deleted as it will be done when DAG finishes running.
+    Saves report with relocated outputs as "workflow_report.json"
+    to "outputs_folder". Maps outputs from "workflow" back to normal
+    (from step_id_step_out to workflow output) and filters "job_data"
+    based on them (combining items from "job_data" into a list based on
+    "outputSource" if it was a list). "cwl_args" can be used to update
+    default parameters used for loading and runtime contexts.
     """
 
     cwl_args = {} if cwl_args is None else cwl_args
-    remove_tmp_folder = True if remove_tmp_folder is None else remove_tmp_folder
 
     default_cwl_args = get_default_cwl_args(cwl_args)
 
@@ -457,12 +503,11 @@ def relocate_outputs(
         else:
             filtered_job_data[output_id] = collected_job_items[0]
 
-    # Outputs will be always copied, because source_directories=[]
     runtime_context = RuntimeContext(default_cwl_args)
     relocated_job_data = relocateOutputs(
         outputObj=filtered_job_data,
         destination_path=job_data_copy["outputs_folder"],
-        source_directories=[],                              # use it as a placeholder (shouldn't influence anything)
+        source_directories=[job_data_copy["tmp_folder"]],   # need to set it to tmp_folder otherwise it won't be able to delete tmp data if action is "move"
         action=runtime_context.move_outputs,
         fs_access=runtime_context.make_fs_access(""),
         compute_checksum=runtime_context.compute_checksum,
@@ -477,10 +522,6 @@ def relocate_outputs(
 
     dump_json(relocated_job_data, workflow_report)
 
-    # Clean "tmp_folder"
-    if remove_tmp_folder:
-        shutil.rmtree(job_data_copy["tmp_folder"], ignore_errors=False)
-
     return relocated_job_data, workflow_report
 
 
@@ -531,6 +572,28 @@ def kill_containers(containers):
             logging.error(f"Failed to kill container. \n {err}")
 
 
+def need_to_run(workflow_data, job_data, task_id):
+    """
+    If step selected by task_id has field "when" then decide
+    whether it should be run or not based on the evaluation of
+    this field. Otherwise return True (run the step)
+    """
+
+    selected_step = list(get_items(workflow_data.tool["steps"], task_id))[0][1]
+    if "when" in selected_step:
+        return do_eval(
+            ex=selected_step["when"],
+            jobinput=job_data,
+            requirements=workflow_data.requirements,
+            outdir=None,
+            tmpdir=None,
+            resources={},
+            timeout=120    # harcoded to 120, because default 30 sec might not be enough when running from Docker
+        )
+    else:
+        return True
+
+
 def execute_workflow_step(
     workflow,
     task_id,
@@ -542,7 +605,10 @@ def execute_workflow_step(
     Constructs and executes single step workflow based on the "workflow"
     and "task_id". "cwl_args" can be used to update default parameters
     used for loading and runtime contexts. Exports json file with the
-    execution results.
+    execution results. If the step was evaluated as the one that need to
+    be skipped, the output "skipped" will set to True and the step_report
+    file will include "nulls". This function doesn't remove any temporary
+    data in both success and failure scenarios.
     """
 
     cwl_args = {} if cwl_args is None else cwl_args
@@ -575,26 +641,33 @@ def execute_workflow_step(
         location=workflow_step_path
     )
 
-    _stderr = sys.stderr                               # to trick the logger	
-    sys.stderr = sys.__stderr__
-    step_outputs, step_status = executor(
-        slow_cwl_load(
-            workflow=workflow_step_path,
-            cwl_args=default_cwl_args),
-        job_data,
-        RuntimeContext(default_cwl_args)
+    workflow_data = slow_cwl_load(
+        workflow=workflow_step_path,
+        cwl_args=default_cwl_args
     )
-    sys.stderr = _stderr
+    
+    skipped = True
+    step_outputs = {output_id: None for output_id, _ in get_items(workflow_data.tool["outputs"])}
+    if need_to_run(workflow_data, job_data, task_id):
+        skipped = False
+        _stderr = sys.stderr                               # to trick the logger
+        sys.stderr = sys.__stderr__
+        step_outputs, step_status = executor(
+            workflow_data,
+            job_data,
+            RuntimeContext(default_cwl_args)
+        )
+        sys.stderr = _stderr
 
-    if step_status != "success":
-        raise ValueError("Failed to run workflow step")
+        if step_status != "success":
+            raise ValueError("Failed to run workflow step")
 
-    # To remove "http://commonwl.org/cwltool#generation": 0 (copied from cwltool)
-    visit_class(step_outputs, ("File",), MutationManager().unset_generation)
+        # To remove "http://commonwl.org/cwltool#generation": 0 (copied from cwltool)
+        visit_class(step_outputs, ("File",), MutationManager().unset_generation)
 
     dump_json(step_outputs, step_report)
 
-    return step_outputs, step_report
+    return step_outputs, step_report, skipped
 
 
 def get_temp_folders(task_id, job_data):
@@ -796,9 +869,12 @@ def fast_cwl_step_load(workflow, target_id, cwl_args=None, location=None):
                 step_in_source_with_step_id = step_in_source.replace("/", "_")  # to include both step name and id
 
                 # Check if it should be assumed optional (default field is present)
+                # Also if we see "pickValue" we should add "null" too, but it might
+                # be not the best solution, as we are not sure whether all of the
+                # items from "source" should be optional or only some of them.
                 # NOTE: consider also checking if upstream step had scatter, so the
                 # output type should become array based on the scatter parameters
-                if "default" in step_in:
+                if "default" in step_in or "pickValue" in step_in:
                     upstream_step_output_type = ["null", upstream_step_output["type"]]
                 else:
                     upstream_step_output_type = upstream_step_output["type"]
